#include "physics.h"

/** VECTORS. This is the bread and butter to be used by
   the game engine. However, the game engine need only call
   vectorTranslate and vectorRotate to be used, everything else
   should be considered internal to physics.c
   (hence everything else is static or inline). */

/* Find the cross product between two vectors. Used in pitch rotations. */
static int vCrossProduct(GLdouble * dst, GLdouble * u, GLdouble * v);

/* Normalizes a vector, pretty straightforward, but needs to be a function. */
static int vNormalize(GLdouble * dst);

/* Finds the magnitude of a vector. */
inline double vMagnitude(GLdouble * u);


/** QUATERNIONS. None of this should be used by the game engine,
   instead the game engine should make calls to vRotate(). */

/* Find q', where if q = (w, v) then q' = (w, -v). */
static int qConjugate(GLdouble * dst, GLdouble * q);

/* Multiply two quaternions. This involves a lot of arithmetic. */
static int qMultiply(GLdouble * dst, GLdouble * q, GLdouble * p);

/* Normalize a quaternion. */
static int qNormalize(GLdouble * q);

/* Find the magnitude of a quaternion. */
inline double qMagnitude(GLdouble * q);

/* Convert a vector into a quaternion where q = (0, v) */
static int qVector2Quat(GLdouble * dst, GLdouble * v);

/* Convert a quaternion back into a vector by dropping the real part. */
static int qQuat2Vector(GLdouble * dst, GLdouble * q);


static int vCrossProduct(GLdouble * dst, GLdouble * u, GLdouble * v)
{
  if (dst == NULL || u == NULL || v == NULL)
    { fprintf(stderr, "vCrossProduct received a null input.\n"); return(1); }
  dst[0] = u[1] * v[2] - u[2] * v[1];
  dst[1] = u[2] * v[0] - u[0] * v[2];
  dst[2] = u[0] * v[1] - u[1] * v[0];
  return(0);
}

static int vNormalize(GLdouble * dst)
{
  if (dst == NULL)
    { fprintf(stderr, "vNormalize received a null input.\n"); return(1); }
  double d = vMagnitude(dst);
  if (d < 0.0f)
  {
    fprintf(stderr, "vNormalize: Unable to normalize null vector.\n");
    return(1);
  }
  else if (d == 0.0f)
  {
    fprintf(stderr, "vNormalize: Unable to normalize zero vector.\n");
    return(1);
  }
  else
  {
    dst[0] /= d;
    dst[1] /= d;
    dst[2] /= d;
  }
  return(0);
}

inline double vMagnitude(GLdouble * u)
{
  if (u == NULL)
    { fprintf(stderr, "vMagnitude received a null input.\n"); return(-1.0f); }
  return(
    sqrt(
      (u[0] * u[0]) +
      (u[1] * u[1]) +
      (u[2] * u[2])
    )
  );
}

static int qConjugate(GLdouble * dst, GLdouble * q)
{
  if (dst == NULL || q == NULL)
    { fprintf(stderr, "qConjugate received a null input.\n"); return(1); }
  dst[0] = -q[0];
  dst[1] = -q[1];
  dst[2] = -q[2];
  return(0);
}

static int qMultiply(GLdouble * dst, GLdouble * a, GLdouble * b)
{
  if (dst == NULL || a == NULL || b == NULL)
    { fprintf(stderr, "qMultiply received a null input.\n"); return(1); }
  dst[0] = a[0] * b[0] - a[1] * b[1] - a[2] * b[2] - a[3] * b[3];
  dst[1] = a[0] * b[1] + a[1] * b[0] + a[2] * b[3] - a[3] * b[2];
  dst[2] = a[0] * b[2] - a[1] * b[3] + a[2] * b[0] + a[3] * b[1];
  dst[3] = a[0] * b[3] + a[1] * b[2] - a[2] * b[1] + a[3] * b[0];
  return(0);
}

static int qNormalize(GLdouble * q)
{
  if (q == NULL)
  {
    fprintf(stderr, "qNormalize received a null input.\n");
    return(1);
  }
  double d = qMagnitude(q);
  if (d < 0.0f)
  {
    fprintf(stderr, "qNormalize: Unable to normalize null quaternion.\n");
    return(1);
  }
  else if (d == 0.0f)
  {
    fprintf(stderr, "qNormalize: Unable to normalize zero quaternion.\n");
    return(1);
  }
  else
  {
    q[0] /= d;
    q[1] /= d;
    q[2] /= d;
    q[3] /= d;
  }
  return(0);
}

inline double qMagnitude(GLdouble * q)
{
  if (q == NULL)
  {
    fprintf(stderr, "qMagnitude received a null input.\n");
    return(-1.0f);
  }
  return(
    sqrt(
      (q[0] * q[0]) +
      (q[1] * q[1]) +
      (q[2] * q[2]) +
      (q[3] * q[3])
    )
  );
}

static int qVector2Quat(GLdouble * dst, GLdouble * v)
{
  if (dst == NULL || v == NULL)
    { fprintf(stderr, "qVector2Quat received a null input.\n"); return(1); }
  dst[0] = 0;
  dst[1] = v[0];
  dst[2] = v[1];
  dst[3] = v[2];
  return(0);
}

static int qQuat2Vector(GLdouble * dst, GLdouble * q)
{
  if (dst == NULL || q == NULL)
    { fprintf(stderr, "qQuat2Vector received a null input.\n"); return(1); }
  qNormalize(q);
  double angle = 2 * acos(q[0]);
  double sine = sin(angle / 2);
  if (sine == 0.0f)
  {
    fprintf(stderr, "qQuat2Vector: prevented accidental divide by zero.\n");
    return(1);
  }
  dst[0] = q[1] / sine;
  dst[1] = q[2] / sine;
  dst[2] = q[3] / sine;
  return(0);
}

int vectorRotate(GLdouble * forward3d, GLdouble * up3d, char * direction, const double angle)
{
  if (forward3d == NULL || up3d == NULL || direction == NULL)
    { fprintf(stderr, "vectorRotate received a null input.\n"); return(1); }
  GLdouble * q = calloc(sizeof(GLdouble), 4); // Rotation quaternion (this is the axis to rotate around)
  GLdouble * q_ = calloc(sizeof(GLdouble), 4); // It's conjugate
  GLdouble * p = calloc(sizeof(GLdouble), 4); // Initial rotation quaterion
  GLdouble * pNew = calloc(sizeof(GLdouble), 4); // Final rotation quaternion
  GLdouble * pTemp = calloc(sizeof(GLdouble), 4); // Temporary rotation quaternion to help 3-step process
  GLdouble * cross; // Cross product of forward3d and up3d (used for pitch).
  q[0] = cos(angle / 2);
  q[1] = q[2] = q[3] = sin(angle / 2);
  switch(direction[0])
  {
    case 'p': // changing pitch has to modify both the up3d and forward3d.
      cross = calloc(sizeof(GLdouble), 3);
      vCrossProduct(cross, forward3d, up3d);
      q[1] *= cross[0];
      q[2] *= cross[1];
      q[3] *= cross[2];
      qConjugate(q_, q);
      
      // This is all forward3d rotation
      qVector2Quat(p, forward3d);
      qMultiply(pTemp, q, p);
      qMultiply(pNew, pTemp, q_);
      qQuat2Vector(forward3d, pNew);
      
      // Now on to up3d rotation
      qVector2Quat(p, up3d);
      qMultiply(pTemp, q, p);
      qMultiply(pNew, pTemp, q_);
      qQuat2Vector(up3d, pNew);
      
      free(cross);
      break;
    case 'r':
      q[1] *= up3d[0];
      q[2] *= up3d[1];
      q[3] *= up3d[2];
      qConjugate(q_, q);
      qVector2Quat(p, forward3d);
      qMultiply(pTemp, q, p);
      qMultiply(pNew, pTemp, q_);
      qQuat2Vector(up3d, pNew);
      break;
    case 'y':
      q[1] *= forward3d[0];
      q[2] *= forward3d[1];
      q[3] *= forward3d[2];
      qConjugate(q_, q);
      qVector2Quat(p, up3d);
      qMultiply(pTemp, q, p);
      qMultiply(pNew, pTemp, q_);
      qQuat2Vector(forward3d, pNew);
      break;
    default:
      fprintf(stderr, "vectorRotate received an input it didn't expect: %s\n", direction);
      break;
  }
  qQuat2Vector(forward3d, pNew);
  free(q);
  free(q_);
  free(p);
  free(pNew);
  free(pTemp);
  vNormalize(forward3d);
  vNormalize(up3d);
  return(0);
}

int vectorTranslate(GLdouble * location3d, GLdouble * forward3d, GLdouble * up3d, const GLdouble radius, char * direction)
{
  if (location3d == NULL || forward3d == NULL || up3d == NULL || direction == NULL)
    { fprintf(stderr, "vectorTranslate received a null input.\n"); return(1); }
  if (radius == 0)
    return(0);
  GLdouble * cross;
  switch(direction[0])
  {
    case 'u':
      location3d[0] += up3d[0] * radius;
      location3d[1] += up3d[1] * radius;
      location3d[2] += up3d[2] * radius;
      break;
    case 'd':
      location3d[0] -= up3d[0] * radius;
      location3d[1] -= up3d[1] * radius;
      location3d[2] -= up3d[2] * radius;
      break;
    case 'l':
      cross = calloc(sizeof(GLdouble), 3);
      vCrossProduct(cross, forward3d, up3d);
      location3d[0] += cross[0] * radius;
      location3d[1] += cross[1] * radius;
      location3d[2] += cross[2] * radius;
      free(cross);
      break;
    case 'r':
      cross = calloc(sizeof(GLdouble), 3);
      vCrossProduct(cross, forward3d, up3d);
      location3d[0] -= cross[0] * radius;
      location3d[1] -= cross[1] * radius;
      location3d[2] -= cross[2] * radius;
      free(cross);
      break;
    case 'f':
      location3d[0] += forward3d[0] * radius;
      location3d[1] += forward3d[1] * radius;
      location3d[2] += forward3d[2] * radius;
      break;
    case 'b':
      location3d[0] -= forward3d[0] * radius;
      location3d[1] -= forward3d[1] * radius;
      location3d[2] -= forward3d[2] * radius;
      break;
    default:
      fprintf(stderr, "vectorRotate received an input it didn't expect: %s\n", direction);
      break;
  }
  return(0);
}
